/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "hdfs.h"

#include "fuse_dfs.h"
#include "fuse_connect.h"
#include "fuse_users.h" 

#include <search.h>

int htable_init = 0;

typedef struct connection_ref_s {
  hdfsFS fs;
  int count;
} connection_ref_t;

struct hsearch_data *user_fs_table;

#if PERMS

pthread_mutex_t connect_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * Connects to the NN as the current user/group according to FUSE
 *
 */

void doConnectAsUser(const char *hostname, int port, hdfsFS *origFS) {

  if (htable_init == 0) {
    user_fs_table = calloc(1, sizeof(struct hsearch_data));
    if (hcreate_r(16*1024, user_fs_table) != 0) {
      htable_init = 1;
    } else {
      ERROR("Unable to initialize connection table");
      exit(1);
    }
  }

  uid_t uid = fuse_get_context()->uid;

  hdfsFS fs = NULL;

  char *user = getUsername(uid);
  if (NULL == user) {
    goto done;
  }

  pthread_mutex_lock(&connect_mutex);

  ENTRY e, *ep, e_new;
  int hret = 0;
  e.key = user;
  hret = hsearch_r(e, FIND, &ep, user_fs_table);
  if ((hret != 0) && (ep != 0)) {
    connection_ref_t *ref = (connection_ref_t*)ep->data;
    if (ref == NULL) {
      ERROR("Retrieved an invalid object");
      goto done;
    }
    ref->count++;
    fs = ref->fs;
  } else {
    fs = hdfsConnectAsUserNewInstance(hostname, port, user);
    if (fs == NULL) {
      ERROR("Unable to instantiate a filesystem for %s", user);
      goto done;
    }
    e_new.key = strdup(user);
    if (e_new.key == NULL) {
      ERROR("Unable to create username; FUSE is likely out of memory");
      goto done;
    }
    connection_ref_t *ref = calloc(1, sizeof(connection_ref_t));
    if (ref != NULL) {
      ref->fs = fs;
      ref->count = 1;
      e_new.data = (void*)ref;
      if ((hsearch_r(e_new, ENTER, &ep, user_fs_table) == 0) && (ep == NULL)) {
        ERROR("Unable to save connection due to full hash table.  FUSE may become unstable");
      }
    } else {
      ERROR("Unable to allocate new reference");
    }
  }

done:

  pthread_mutex_unlock(&connect_mutex);

  if (user) 
    free(user);

  *origFS = fs;
}

#else
hdfsFS defaultFS = NULL;
hdfsFS doConnectAsUser(const char *hostname, int port) {
  if (!defaultFS)
    defaultFS = hdfsConnect(hostname, port);

  return defaultFS;
}

#endif


/**
 * Connects to the NN as the default user.
 * Useful for testing whether the HDFS setup is usable.
 */
hdfsFS doConnect(const char *hostname, int port) {

  hdfsFS fs = NULL;
  fs = hdfsConnectAsUser(hostname, port, "root");
  
  return fs;
}

/**
 * Disconnects the given HDFS file system object.
 * @param fs The configured filesystem handle.
 * @return Returns 0 on success, -1 on error.
 */
int doDisconnect(hdfsFS fs) {
  // This is implemented as a separate function to ease the future transition
  // to FileContext.

  // Currently we never disconnect filesystems, as this causes the JVM to segfault.
  return 0;
/*
  int retval = 0;
  retval = hdfsDisconnect(fs);
  return retval;
*/
}

